use crate::robot::{Error, ROBOT_TIMEOUT};
use log::{error, info};
use prost::Message;
use raas_types::raas::cmd::{Command, Request};
use raas_types::raas::register::BotTypes;
use raas_types::raas::resp::Response;
use raas_types::raas::{recv_raas_msg, send_raas_msg};
use tokio::io::DuplexStream;
use tokio::net::TcpStream;

#[derive(Debug)]
pub struct RobotConnector {
    pub bot_id: u32,
    pub bot_tags: Vec<BotTypes>,

    tcp_stream: TcpStream,
    msg_stream: DuplexStream,

    is_running: bool,
}

impl RobotConnector {
    pub fn new(
        bot_id: u32,
        bot_tags: Vec<BotTypes>,
        tcp_stream: TcpStream,
        msg_stream: DuplexStream,
    ) -> Self {
        Self {
            bot_id,
            bot_tags,
            tcp_stream,
            msg_stream,
            is_running: false,
        }
    }

    async fn wait_for_request(&mut self) -> Result<Request, Error> {
        let request = recv_raas_msg(&mut self.msg_stream).await?;

        let request = Request::decode(request.msg.as_slice())?;

        Ok(request)
    }

    async fn send_command(&mut self, request: Request) -> Result<(), Error> {
        let cmd = Command {
            id: self.bot_id,
            request: Some(request),
        };

        let mut msg = Vec::new();

        cmd.encode(&mut msg).unwrap();
        send_raas_msg(&mut self.tcp_stream, msg).await?;

        Ok(())
    }

    async fn get_response(&mut self) -> Result<Response, Error> {
        let recv_timeout = tokio::time::timeout(ROBOT_TIMEOUT, recv_raas_msg(&mut self.tcp_stream));
        let recv = recv_timeout.await??;

        let resp = Response::decode(recv.msg.as_slice()).unwrap();

        info!("Worker (bot_id={}) got resp", self.bot_id);

        Ok(resp)
    }

    async fn respond_to_request(&mut self, resp: Response) -> Result<(), Error> {
        let mut msg = Vec::new();
        resp.encode(&mut msg)?;

        send_raas_msg(&mut self.msg_stream, msg).await?;

        Ok(())
    }

    pub async fn worker(&mut self) {
        self.is_running = true;

        loop {
            if let Err(res) = self.handle_request().await {
                error!("Closing robot connection: {}", res);
                break;
            }
        }

        error!("Exiting Worker (bot_id={})", self.bot_id);
        self.is_running = false;
    }

    async fn handle_request(&mut self) -> Result<(), Error> {
        info!("Worker (bot_id={}) is waiting for requests", self.bot_id);
        let request = match self.wait_for_request().await {
            Ok(r) => r,
            Err(err) => {
                error!(
                    "Worker (bot_id={}) failed to get request: {:?}",
                    self.bot_id, err
                );
                return Err(err);
            }
        };

        match self.send_command(request).await {
            Ok(_) => {}
            Err(err) => {
                error!(
                    "Worker (bot_id={}) failed to send command to bot: {:?}",
                    self.bot_id, err
                );
                return Err(err);
            }
        }

        let resp = match self.get_response().await {
            Ok(r) => r,
            Err(err) => {
                error!(
                    "Worker (bot_id={}) failed to get response from bot: {:?}",
                    self.bot_id, err
                );
                return Err(err);
            }
        };

        match self.respond_to_request(resp).await {
            Ok(_) => Ok(()),
            Err(err) => {
                error!(
                    "Worker (bot_id={}) failed to send response: {:?}",
                    self.bot_id, err
                );
                Err(err)
            }
        }
    }
}
